/*
 * Copyright (C) 2001-2007 by egnite Software GmbH
 * Copyright (C) 2008-2010 by egnite GmbH
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of the copyright holders nor the names of
 *    contributors may be used to endorse or promote products derived
 *    from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS
 * FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE
 * COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * For additional information see http://www.ethernut.de/
 *
 */

/*!
 * \file httpcgi.c
 * \brief HTTP CGI interface.
 *
 * \verbatim
 * $Id$
 * \endverbatim
 */

#include "config.h" /* Must be first. */

#if USE_HTTPSERVER

#include <stdlib.h>
#include <string.h>
//#include <nut_io.h>
#include <fcntl.h>
#include <time.h>

#include <dev/board.h>
#include <dev/urom.h>
#include <dev/vscodec.h>
//@@MF #include <dev/watchdog.h>

#include <arpa/inet.h>
#include <netinet/tcp.h>
#include <netdb.h>

#include <sys/confos.h>
#include <sys/confnet.h>
#include <sys/thread.h>
#include <sys/timer.h>
#include <sys/heap.h>
#include <sys/bankmem.h>
#include <sys/socket.h>
#include <sys/version.h>

#include <pro/ssi.h>
#include <pro/httpd.h>

#include <memdebug.h>

#include "logmsg.h"
#include "utils.h"
#include "xmlutils.h"
#include "favlist.h"
#include "proxy.h"
#include "webradio.h"
#include "shoutcast.h"
#include "httpserv.h"

#define CGI_PARAM_KEY_VALUE     "value"
#define CGI_PARAM_IDX_VALUE     0

#define CGI_PARAM_KEY_FUNC      "func"
#define CGI_PARAM_IDX_FUNC      1

#define CGI_PARAM_KEY_HOSTNAME  "hostname"
#define CGI_PARAM_IDX_HOSTNAME  2

#define CGI_PARAM_KEY_PROXY     "proxy"
#define CGI_PARAM_IDX_PROXY     3

#define CGI_PARAM_KEY_DHCP      "dhcp"
#define CGI_PARAM_IDX_DHCP      4

#define CGI_PARAM_KEY_IP        "ip"
#define CGI_PARAM_IDX_IP        5

#define CGI_PARAM_KEY_MASK      "mask"
#define CGI_PARAM_IDX_MASK      6

#define CGI_PARAM_KEY_GATE      "gateway"
#define CGI_PARAM_IDX_GATE      7

#define CGI_PARAM_KEY_DNS0      "dns0"
#define CGI_PARAM_IDX_DNS0      8

#define CGI_PARAM_KEY_DNS1      "dns1"
#define CGI_PARAM_IDX_DNS1      9

#define CGI_PARAM_MAX_COUNT     10


static char *param_keywords[] = {
    CGI_PARAM_KEY_VALUE,
    CGI_PARAM_KEY_FUNC,
    CGI_PARAM_KEY_HOSTNAME,
    CGI_PARAM_KEY_PROXY,
    CGI_PARAM_KEY_DHCP,
    CGI_PARAM_KEY_IP,
    CGI_PARAM_KEY_MASK,
    CGI_PARAM_KEY_GATE,
    CGI_PARAM_KEY_DNS0,
    CGI_PARAM_KEY_DNS1
};

#define CGI_PARAM_UPD_OS        0x01
#define CGI_PARAM_UPD_NET       0x02
#define CGI_PARAM_UPD_RADIO     0x04
#define CGI_PARAM_UPD_BOOT      0x80


/*!
 * \brief Dynamically create the station favorites list or edit.
 */
static int CgiStationFavorites(FILE * stream, REQUEST * req)
{
    int i;
    int count;
    int action = 0;
    int idx = 0;

    /* Parsing the parameters:
    ** index=edit edit an existing entry.
    ** 0=add add a new entry.
    **/
    count = NutHttpGetParameterCount(req);
    while (count--) {
        char *val = NutHttpGetParameterValue(req, count);
        if (strcmp(val, "edit") == 0) {
            /* Editing an existing entry. */
            action = 1;
            idx = atoi(NutHttpGetParameterName(req, count));
        } else if (strcmp(val, "add") == 0) {
            /* Adding a new entry. */
            action = 2;
        }
    }

    /*
     * Edit existing or add new.
     */
    if (action) {
        fprintf(stream,
            "<tr class=\"tr0\">\r\n" /* First row. */
            "<td>Station Name</td>\r\n" /* Text field column. */
            " <td align=\"left\">\r\n" /* Edit field column. */
            "  <input type=\"hidden\" name=\"%d\" value=\"%s\">\r\n" /* Hidden index. */
            "  <input type=\"text\" name=\"sname\" value=\"%s\" class=\"inputl\">\r\n" /* Edit field. */
            " </td>\r\n" /* */
            "</tr>\r\n", /* */
            idx, action == 1 ? "saveedit" : "savenew", /* Hidden name and value. */
            action == 1 ? favlist[idx].rs_name : ""); /* Edit field contents. */
        fputs("<tr class=\"tr1\">\r\n" /* Second row. */
              " <td>Station URLs</td>\r\n" /* Text field column. */
              " <td align=\"left\">" /* Edit field column. */
              "  <textarea class=\"inputl\" name=\"surl\" style=\"height:100;\">", stream);
        if (action == 1) {
            /* Initial edit field entry. */
            for (i = 0; i < favlist[idx].rs_streams; i++) {
                fprintf(stream, "%s\r\n", favlist[idx].rs_uri[i]);
            }
        }
        fputs("  </textarea>\r\n" /* */
              " </td>\r\n" /* */
              "</tr>\r\n" /* */
              "<tr class=\"tr0\">\r\n" /* Third row. */
              " <td></td>\r\n" /* First column. */
              " <td align=\"right\">" /* Second column. */
              "  <input type=\"submit\" name=\"func\" value=\"Apply\" class=\"submit\">" /* */
              " </td>\r\n" /* */
              "</tr>\r\n", stream);
    }

    /*
     * List all entries.
     */
    else {
        int row = 0;

        for (i = LAST_FAVORITE + 1; i < MAX_FAVORITES; i++, row++) {
            if (favlist[i].rs_name == NULL) {
                continue;
            }

            fprintf(stream,
                "<tr class=\"tr%d\">\r\n" /* Row start. */
                " <td>%s</td>\r\n" /* Name column. */
                " <td><input type=\"submit\" name=\"%d\" value=\"edit\" class=\"submit\"></td>\r\n" /* Edit */
                " <td><input type=\"submit\" name=\"%d\" value=\"delete\" class=\"submit\" onClick=\"return confirm('Delete %s?');\"></td>\r\n"
                " <td><input type=\"submit\" name=\"%d\" value=\"play\" class=\"submit\"></td>\r\n" /* Play */
                "</tr>\r\n" /* Row end. */
                , row & 1, favlist[i].rs_name, i, i, favlist[i].rs_name, i);
        }
        fprintf(stream,
            "<tr class=\"tr%d\">\r\n" /* Last row. */
            " <td colspan=\"4\" align=\"right\"><input type=\"submit\" name=\"0\" value=\"add\" class=\"submit\"></td>\r\n"
            "</tr>\r\n", row & 1);
    }
    return 0;
}

const char strGenre[] = "genre";
const char strSearch[] = "search";
const char strAction[] = "action";

static int CgiShoutCastSearch(FILE * stream, REQUEST * req)
{
    int i;
    int count;
    char *name;
	char *getsearch = NULL;
	char *action = NULL;

    count = NutHttpGetParameterCount(req);
    for (i = 0; i < count; i++) {
        name = NutHttpGetParameterName(req, i);
		if (strcmp(name, strSearch) == 0) {
			getsearch = NutHttpGetParameterValue(req, i);
		}
		else if (strcmp(name, "btn") == 0) {
			action = NutHttpGetParameterValue(req, i);
		}
    }
    if (getsearch) {
		fputs(getsearch, stream);
    }
    fflush(stream);

	return 0;
}

static int CgiShoutCastGenres(FILE * stream, REQUEST * req)
{
    int i;
    int count;
    char *name;
    char *genre;
    char *selected = NULL;

    /* Check, if we had a previous selection. */
    count = NutHttpGetParameterCount(req);
    for (i = 0; i < count; i++) {
        name = NutHttpGetParameterName(req, i);
        if (strcmp(name, strGenre) == 0) {
            selected = NutHttpGetParameterValue(req, i);
			break;
        }
    }
    if (strcmp(selected, "Any") == 0) {
        selected = NULL;
    }

    fputs("<option", stream);
	if (selected == NULL) {
        fputs(" selected", stream);
	}
    fprintf(stream, ">Any</option>\r\n");

    for (i = 0;; i++) {
        genre = ShoutCastGetGenre(i);
        if (genre == NULL) {
            break;
        }
        fputs("<option", stream);
        if (selected && strcmp(genre, selected) == 0) {
            fputs(" selected", stream);
            selected = NULL;
        }
        fprintf(stream, ">%s</option>\r\n", genre);
    }
    fflush(stream);

    return 0;
}

static int CgiShoutCastStations(FILE * stream, REQUEST * req)
{
    int count;
    char *name = NULL;
    char *genre = NULL;
	char *search = NULL;
	char *action = NULL;
    int i;

	LogDbgMsg(LOG_SHOUTCAST, "CSCS:\r\n");
    count = NutHttpGetParameterCount(req);
    for (i = 0; i < count; i++) {
        name = NutHttpGetParameterName(req, i);
        if (strcmp(name, strGenre) == 0) {
            genre = NutHttpGetParameterValue(req, i);
			LogDbgMsg(LOG_SHOUTCAST, ">%s=%s\r\n", name, genre);
        }
		else if (strcmp(name, strSearch) == 0) {
			search = NutHttpGetParameterValue(req, i);
			LogDbgMsg(LOG_SHOUTCAST, ">%s=%s\r\n", name, search);
		}
		else if (strcmp(name, strAction) == 0) {
			action = NutHttpGetParameterValue(req, i);
			LogDbgMsg(LOG_SHOUTCAST, ">%s=%s\r\n", name, action);
		}
    }
    for (i = 0;; i++) {
		if( search && action && (strcmp(action, "Search") == 0)) {
			/* Text search requested */
            name = ShoutCastGetStationNameBySearch(search, i);
		}
		else if( genre) {
			/* Search for genre */
            name = ShoutCastGetStationNameByGenre(genre, i);
        }
        else {
			/* Empty request (Initial page load) */
            name = ShoutCastGetStationNameByGenre("Euro", i);
        }

        if (name == NULL) {
            break;
        }

        if (i & 1) {
            fputs("<tr class=\"tr0\">\r\n", stream);
        } else {
            fputs("<tr class=\"tr1\">\r\n", stream);
        }
        fprintf(stream, "<td>%s</td>\r\n", name);
        fprintf(stream, "<td><input type=\"submit\" name=\"%d\" value=\"play\" class=\"submit\"></td>\r\n", i);
        fprintf(stream, "<td><input type=\"submit\" name=\"%d\" value=\"add to favorites\" class=\"submit\"></td>\r\n", i);
        fputs("</tr>\r\n", stream);
    }
    fflush(stream);

    return 0;
}

static int CgiShoutCastControl(FILE * stream, REQUEST * req)
{
    int count;
    char *name;
    char *value;
    int action = 0;
    int idx = 0;

    count = NutHttpGetParameterCount(req);
    while (count--) {
        value = NutHttpGetParameterValue(req, count);
        if (strcmp(value, "add to favorites") == 0) {
            action = 1;
        }
        if (strcmp(value, "play") == 0) {
            action = 2;
        }
        if (action) {
            if ((name = NutHttpGetParameterName(req, count)) == NULL) {
                action = 0;
            } else {
                idx = atoi(name);
                break;
            }
        }
    }
    if (action == 1) {
        ShoutCastAddStation(idx, MAX_FAVORITES);
        webradio.wr_favupd = 10;
    } else if (action == 2) {
        ShoutCastAddStation(idx, TOP_FAVORITE);
    }
    return 0;
}

static int CgiFavoritesControl(FILE * stream, REQUEST * req)
{
    int count;
    char *name;
    char *value;
    int action = 0;
    int idx = 0;
    char *sname = NULL;
    char *surl = NULL;

    count = NutHttpGetParameterCount(req);
    while (count--) {
        name = NutHttpGetParameterName(req, count);
        value = NutHttpGetParameterValue(req, count);
        if (strcmp(name, "sname") == 0) {
            sname = value;
        } else if (strcmp(name, "surl") == 0) {
            surl = value;
        } else if (strcmp(value, "saveedit") == 0) {
            action = 1;
            idx = atoi(name);
        } else if (strcmp(value, "delete") == 0) {
            action = 2;
            idx = atoi(name);
        } else if (strcmp(value, "play") == 0) {
            action = 3;
            idx = atoi(name);
        } else if (strcmp(value, "savenew") == 0) {
            action = 4;
        }
    }
    if (action == 1) {
        /* Save existing. */
        idx = FavListSet(idx, sname, NULL);
        AddStationUris(idx, surl);
        webradio.wr_favupd = 2;
    } else if (action == 2) {
        FavListSet(idx, NULL, NULL);
        webradio.wr_favupd = 20;
    } else if (action == 3) {
        FavListCopy(idx, TOP_FAVORITE);
    } else if (action == 4) {
        /* Add new. */
        idx = FavListSet(MAX_FAVORITES, sname, NULL);
        AddStationUris(idx, surl);
        webradio.wr_favupd = 2;
    }
    return 0;
}

static uint_fast8_t GetSettingsIndex(char *key)
{
    uint_fast8_t rc;

    for (rc = 0; rc < CGI_PARAM_MAX_COUNT; rc++) {
        if (strcmp(key, param_keywords[rc]) == 0) {
            break;
        }
    }
    return rc;
}

/*!
 * \brief CGI callback for basic parameter settings.
 */
static int CgiBasicSettings(FILE * stream, REQUEST * req)
{
    int count;
    int i;
    char *sp;
    uint32_t addr;
    uint_fast8_t update = 0;
    char **valtab;
    uint_fast8_t idx;

    valtab = malloc(CGI_PARAM_MAX_COUNT * sizeof(char *));
    if (valtab == NULL) {
        return -1;
    }
    memset(valtab, 0, CGI_PARAM_MAX_COUNT * sizeof(char *));

    count = NutHttpGetParameterCount(req);
    for (i = 0; i < count; i++) {
        sp = NutHttpGetParameterName(req, i);
        idx = GetSettingsIndex(sp);
        if (idx < CGI_PARAM_MAX_COUNT) {
            valtab[idx] = NutHttpGetParameterValue(req, i);
        }
    }
    if ((sp = valtab[CGI_PARAM_IDX_VALUE]) != NULL) {
        idx = GetSettingsIndex(sp);

        sp = NULL;
        addr = (uint32_t)-1;
        switch(idx) {
        case CGI_PARAM_IDX_HOSTNAME:
            sp = confos.hostname;
            break;
        case CGI_PARAM_IDX_PROXY:
            if (proxy.proxy_port) {
                fputs(proxy.proxy_host, stream);
                if (proxy.proxy_port != 80) {
                    fprintf(stream, ":%u", proxy.proxy_port);
                }
            }
            break;
        case CGI_PARAM_IDX_DHCP:
            if (confnet.cdn_cip_addr == 0) {
                sp = " checked=\"checked\"";
            }
            break;
        case CGI_PARAM_IDX_IP:
            if (confnet.cdn_cip_addr) {
                addr = confnet.cdn_cip_addr;
            } else {
                addr = confnet.cdn_ip_addr;
            }
            break;
        case CGI_PARAM_IDX_MASK:
            addr = confnet.cdn_ip_mask;
            break;
        case CGI_PARAM_IDX_GATE:
            addr = confnet.cdn_gateway;
            break;
        case CGI_PARAM_IDX_DNS0:
            NutDnsGetConfig2(NULL, NULL, &addr, NULL);
            break;
        case CGI_PARAM_IDX_DNS1:
            NutDnsGetConfig2(NULL, NULL, NULL, &addr);
            break;
        }
        if (addr != (uint32_t)-1) {
            sp = inet_ntoa(addr);
        }
        if (sp) {
            fputs(sp, stream);
        }
    }
    else if ((sp = valtab[CGI_PARAM_IDX_FUNC]) != NULL) {
        if (strcmp(sp, "Erase") == 0) {
            uint8_t mac[6];

            memcpy(mac, confnet.cdn_mac, sizeof(mac));
            memset(&confnet, 0, sizeof(confnet));
            memcpy(confnet.cdn_mac, mac, sizeof(confnet.cdn_mac));

            ConfigResetFactory();
            FavListResetFactory();
            if (cfgFile) {
                FavListSave(cfgFile);
            }
            update |= CGI_PARAM_UPD_NET | CGI_PARAM_UPD_RADIO;
            webradio.wr_reboot = 2;
        }
        else if (strcmp(sp, "Apply") == 0) {
            if ((sp = valtab[CGI_PARAM_IDX_HOSTNAME]) != NULL) {
                strncpy(confos.hostname, sp, sizeof(confos.hostname) - 1);
                update |= CGI_PARAM_UPD_OS;
            }
            if ((sp = valtab[CGI_PARAM_IDX_PROXY]) != NULL) {
                while (*sp && *sp <= ' ') {
                    sp++;
                }
                if (strlen(sp)) {
                    strncpy(proxy.proxy_host, sp, sizeof(proxy.proxy_host) - 1);
                    if ((sp = strrchr(proxy.proxy_host, ':')) != NULL) {
                        *sp++ = 0;
                        proxy.proxy_port = atoi(sp);
                    }
                }
                else {
                    memset(&proxy, 0, sizeof(proxy));
                }
                update |= CGI_PARAM_UPD_RADIO;
            }
            if ((sp = valtab[CGI_PARAM_IDX_DHCP]) != NULL) {
                if (strcmp(sp, "on") == 0) {
                    confnet.cdn_cip_addr = 0;
                    update |= CGI_PARAM_UPD_NET | CGI_PARAM_UPD_BOOT;
                } else {
                    for (idx = CGI_PARAM_IDX_IP; idx <= CGI_PARAM_IDX_DNS1; idx++) {
                        if ((sp = valtab[idx]) != NULL) {
                            addr = inet_addr(sp);
                        } else {
                            addr = (uint32_t)-1;
                        }
                        if (addr != (uint32_t)-1) {
                            switch(idx) {
                            case CGI_PARAM_IDX_IP:
                                confnet.cdn_cip_addr = addr;
                                update |= CGI_PARAM_UPD_NET;
                                break;
                            case CGI_PARAM_IDX_MASK:
                                confnet.cdn_ip_mask = addr;
                                update |= CGI_PARAM_UPD_NET;
                                break;
                            case CGI_PARAM_IDX_GATE:
                                confnet.cdn_gateway = addr;
                                update |= CGI_PARAM_UPD_NET;
                                break;
                            case CGI_PARAM_IDX_DNS0:
                                webradio.wr_pridns = addr;
                                update |= CGI_PARAM_UPD_RADIO;
                                break;
                            case CGI_PARAM_IDX_DNS1:
                                webradio.wr_secdns = addr;
                                update |= CGI_PARAM_UPD_RADIO;
                                break;
                            }
                            update |= CGI_PARAM_UPD_BOOT;
                        }
                    }
                }
            }
        }

        if (update & CGI_PARAM_UPD_OS) {
            LogMsg(LOG_HTTPD, "Saving OS\r\n");
            NutSaveConfig();
        }
        if (update & CGI_PARAM_UPD_NET) {
            LogMsg(LOG_HTTPD, "Saving Net\r\n");
            NutNetSaveConfig();
        }
        if (cfgFile) {
            if (update & CGI_PARAM_UPD_RADIO) {
                ConfigSave(cfgFile);
            }
            if (update & CGI_PARAM_UPD_BOOT) {
                LogMsg(LOG_HTTPD, "Reboot\r\n");
                webradio.wr_reboot = 3;
            }
        }
    }
    free(valtab);
    return 0;
}

static void PrintMHz(FILE *stream, uint32_t freq)
{
    fprintf(stream, "%lu.%04lu MHz", freq / 1000000UL, (freq - (freq / 1000000UL) * 1000000UL) / 100);
}

/*!
 * \brief Print variable contents.
 */
static int CgiVars(FILE * stream, REQUEST * req)
{
    int count;
    char *name;
    char *cgivar = NULL;
    int i;

    count = NutHttpGetParameterCount(req);
    for (i = 0; i < count; i++) {
        name = NutHttpGetParameterName(req, i);
        if (strcmp(name, "var") == 0) {
            cgivar = NutHttpGetParameterValue(req, i);
            break;
        }
    }

    if (cgivar == NULL) {
        fputs("Error", stream);
    } else if (strcmp(cgivar, "MetaTitle") == 0 || strcmp(cgivar, "StreamTitle") == 0) {
        if (webradio.wr_rip) {
            SHOUTCASTINFO *sci = (SHOUTCASTINFO *) webradio.wr_rip->ri_bcast;
            if (sci && sci->sci_metatitle) {
                fputs(sci->sci_metatitle, stream);
            }
        }
    } else if (strcmp(cgivar, "CPUClock") == 0) {
        PrintMHz(stream, NutGetCpuClock());
    } else if (strcmp(cgivar, "MasterClock") == 0) {
        PrintMHz(stream, NutClockGet(NUT_HWCLK_PERIPHERAL));
    } else if (strcmp(cgivar, "TotalMemoryUsed") == 0) {
        size_t mem = NUTMEM_SIZE;
#if NUTXMEM_SIZE
        mem += NUTXMEM_SIZE;
#endif
        fprintf(stream, "%u kBytes", (unsigned int) (mem - NutHeapAvailable()) / 1024);
    } else if (strcmp(cgivar, "StationName") == 0) {
        if (webradio.wr_sip) {
            fputs(webradio.wr_sip->si_name, stream);
        }
    } else if (strcmp(cgivar, "StationGenre") == 0) {
        if (webradio.wr_sip && webradio.wr_sip->si_genre) {
            fputs(webradio.wr_sip->si_genre, stream);
        }
    } else if (strcmp(cgivar, "StationBitrate") == 0) {
        if (webradio.wr_sip) {
            fprintf(stream, "%u kBit", webradio.wr_sip->si_bitrate);
        }
    } else if (strcmp(cgivar, "BufferContents") == 0) {
        fprintf(stream, "%lu Bytes", NutSegBufUsed());
    } else if (strcmp(cgivar, "StreamURL") == 0) {
        if (webradio.wr_rip) {
            SHOUTCASTINFO *sci = (SHOUTCASTINFO *) webradio.wr_rip->ri_bcast;
            if (sci->sci_metaurl) {
                fputs(sci->sci_metaurl, stream);
            }
        }
    } else if (strcmp(cgivar, "MetadataInterval") == 0) {
        if (webradio.wr_rip) {
            SHOUTCASTINFO *sci = (SHOUTCASTINFO *) webradio.wr_rip->ri_bcast;
            fprintf(stream, "%lu Bytes", sci->sci_metaint);
        }
    } else if (strcmp(cgivar, "NutVersion") == 0) {
        fputs(NutVersionString(), stream);
    } else if (strcmp(cgivar, "AppVersion") == 0) {
        fputs(VERSION, stream);
    } else if (strcmp(cgivar, "Hardware") == 0) {
        fputs(BOARD_NAME, stream);
    } else {
        fputs("Unknown", stream);
    }
    fflush(stream);
    return 0;
}

/*!
 * \brief HTTP Server Start.
 */
int RegisterCgis(void)
{
    /*
     * Register SSI handler
     */
    NutRegisterSsi();

    /*
     * Register our CGIs.
     */
    NutRegisterCgi("f_control.cgi", CgiFavoritesControl);
    NutRegisterCgi("favorites.cgi", CgiStationFavorites);
    NutRegisterCgi("genres.cgi", CgiShoutCastGenres);
	NutRegisterCgi("search.cgi", CgiShoutCastSearch);
    NutRegisterCgi("stations.cgi", CgiShoutCastStations);
    NutRegisterCgi("sc_control.cgi", CgiShoutCastControl);
    NutRegisterCgi("vars.cgi", CgiVars);
    NutRegisterCgi("settings.cgi", CgiBasicSettings);

    return 0;
}

#endif /* USE_HTTPSERVER */
